# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from outputs.png

icon_outputs = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
               b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAN1wAADdcBQiibeAAA'\
               b'CuNJREFUaIHVmltsXcd1hr+Z2ftceZVoi4xFS6KdSlHkooGr'\
               b'JgpyqWQEKFwkqNvaemrhINVDCxfpS4HW7YP8UD+5ShEbiuJL'\
               b'0LhGW0Bx/ZDGdYwiiA2bVS624jqOqBspUbLEm07Ec9vXWdOH'\
               b'fXgkUhR5KJkPXcAAm+SamfXPWutfa8+mYhX57ksvTez6xCe2'\
               b'Gt+jq9xFIZ/H8zy01hhjaDabvPXWW+zbtw+tNSJCkqaEYUi9'\
               b'XqNarXH6zJnxr331q/esttetiLeawq9/Xdm6cWCAOI4xxuD5'\
               b'PlopcrkcuVyORqPB5OQk5XIZAGstJk1J0pT5apVfnTjBlUpl'\
               b'ZD2MB9CrKVQqFcbPnmViYoKJ8XFOjo1Rq9dwOABEhCNHjuCc'\
               b'wzmHOIdYi01T0tQizq2X7UAHHujt7WPrtq3EcUIulyNJEur1'\
               b'Ohs2bGjr+L7fflaLZq+v8dCBB8QJIq59ws4JWpsV5zho6QLr'\
               b'7IHVAYi0jadlj9aLp8Ub77zJbNcGs16yaghdO/lrQ6nFgdLY'\
               b'vuvGedz88Dc9dPjOgip+1hil6mEyOvP9A9O3ZD1r9MDCWOoB'\
               b'EzSvGe5cZnkWPzekweaH/7FYcubAx4e7vrNrpP/ZjWX9xJav'\
               b'HP7UrQLoyAPipM06ywHo/uDdZec5R3vegjRrVb/buG0Pfu7+'\
               b'/vs+OcybPzn7tXd/OXkfX/7mM82q/v7sG4/V1wKgcw9cl8h6'\
               b'SQipOF5s/DJPbamADatGa0H5Pl/Y8xveV76067MPfGbLoTs3'\
               b'qX/Y/OVn7l0LgM5yQDJ+X0hGpZfH7ZY8L0TRIh17WdmwbIKg'\
               b'SdAMERyDg/383t5dg73dxT8/fmJqq/8Hh5+bODX9Or96YvHJ'\
               b'3CqABS8sWLM0hNpAr/2wSH+ppFFVh0FAEEbXwkwrfvu3tvob'\
               b'+sq//4ux6W3lrq6R2e0v/vv0K386s5J9q4YQLITRNRBLWQgg'\
               b'jmNmpqex1q5YviSdVjap6yAICIKYMIwIg5gwiEmsZXhzv/ni'\
               b'79x936fv23Tw7g25x0ceevFGirtOVvUAgBW7mFiW4UfnHKVy'\
               b'mUYQ8/LrP+fyzDzVepWr81XCIOTjD3z9nyQNsHGUS+Lg/iiK'\
               b'iOKk7SSHa4ern/P4ze0b+3tK3l+cmLj6KW//dw+dmr/jNV57'\
               b'MLolACLSYsQWE4ncoFMoFMgXCpw4fZE33znN5uEhujYOUto4'\
               b'hHOOLTvk6yIWsYITy+DQx4jj9DoAZF4WyfJNhM2DJd/3+EJ3'\
               b'ybvLG6/srj78z9+6ePTRD9cMwFqL1jpjIlixQXNOEBHK3SVy'\
               b'+QKe52N8D6M9jDEordFaoxQkseAU7ZxZaFXEZmuICN1lzY4t'\
               b'pRHl7F9fqoTbcg9/54XxmfNv8sYTaecesAKoRR3oTQHgCMOI'\
               b'ZjPCYRAB48AYhRHQBrQGpTQKAbVQ96RF1RkAay1WLNam2NSq'\
               b'uwf9XLngHskpu924Lc8FDz939OLRA5UOc0BQWrcZQ8TeXNlB'\
               b'GEU0myFK+4hoPDF4vuCcRjvXAtAiA7cQ/7RaccnacWtJbYqz'\
               b'ljRNiOMIlTY9L67c74Jwq9HFe+7+wxeOdJwDGQvJTZM4Mz7L'\
               b'8jAIqTdCtJfH4eGU4LQDBaalpvS1Oe33CHHYNDv9KIwJg4hm'\
               b'EFCvN5ifrzJ35SqNep1mEG4UU/5Lv2dwW8c5YDzTrk4rhpA4'\
               b'qtUa5yen2TAg9PQIpbIjX4BcDjzPobXAwuk7hxUhTYU0scRJ'\
               b'QhjGBEFAFEUEzWb2HAYkqSAOcAKS+qRxV8cAsgR2CCsDAAjD'\
               b'iLkrNVKXJ4g05QCKJSgUHJ5n8VrJrFSWV04yEGItSSokFsRp'\
               b'HAqlNcYsJL4CpZ3yyxXtd79skc5CKKsDGbWtFEJZNXY4ScG5'\
               b'bHPPYHwfz/fxPB/f9/A80zZoodIbEVKlsjwTh/UM2hq0NqhW'\
               b'vCXWprFVp1LtH7GR/t7cf//t5VUB2DQliWNMuYtavYHYlOHh'\
               b'YbTSy7YUOEFsgnVZomtt8IyH7xn8nEfO9zMAKgOw8Ma3UMFF'\
               b'BCuZ4ZmOxlpLtVazjUbzB5G13557+9LrcNRCBzRqPA9rhTAM'\
               b'6evrpau7m2Kx2ObpZRCglKVZvcK8Udi4SbN2hUI+R84zeEZj'\
               b'jOKurdvp7t3QMljQrdMXIxgrmNbhNOpVNzN96VK1Ov9vYZR8'\
               b'u3r8hTPX79ZRCDmxxHGMtNy7kgxs7OOhL32Gk2cnCaKAMJwn'\
               b'asR8cH72ZSQFiY3YZPeDf3zgrg0DmxARtLXgwEhmuNaKNI25'\
               b'NDkuM1Mf/rRarT1zdebKK1z+z+bS/TqjUSBXyEO7obv5fcPG'\
               b'vh7+6sAfMTc3x8S5CSYvXKDeaHDwh//yCADdO/pzxdxhrfUj'\
               b'uZyPtRalaAFROKAyN8X5MyevXLgw8UoQREeaY0ePt8y4QTpm'\
               b'IVo8LU5WvGhY9M6gVLt1aBug8xatref75Hy/lbgOrS1pEnP+'\
               b'zPv2zNj747NTlw/HNj0anP7BhzffbQ0AFqrwzUJobm4u43Rr'\
               b'aTab1Gs1mvUGjUY22qI9h3LWmIyNRASlFLX5Cu+/M5qc+OXP'\
               b'X6vNX30+OCevwhvparZ1XInlul5lORkYGEBECIKAsZMn6evv'\
               b'p1AsUCwWl6Nd8TyDMQZrA86OvefePfbjS+fGT/5rHNSejy/9'\
               b'7FQndnUMIEnSa53iKuHz7LPPEscxs3NzfHHv797IVMo4EAmC'\
               b'kLmZKX729o/kg/eO/c/Uh+e/lVyeeRnOh50a3zGA1KaL7j2d'\
               b'ZP1KmqakNuP7NE2xIpRKJQC6u7qWL3jKc07Z+uiPf+h++vaP'\
               b'Khcnz7waBbVvplPHjwMrdIm3AUCsJQpCnHOEYciFC5MMbhpk'\
               b'5J57SJOE4TBsg9m/fz8nxsbIFwpcuDB5I4gro00Z/PSrk+On'\
               b'Upz93yRJXqfy3oqJelsAlFIkSUI+n0dpTblUZnjzMIVCnjiO'\
               b'SZKETzYaxEnS/jYwODREpVJBxKFQSyu2tVM/+S8Lr/ER3P6u'\
               b'CkBrTZIkWLFEQZNCvkAcx1kz6bLQaXz+88RRlHWVSUIUhsRx'\
               b'jBWL0mrZS4CPwviOABitCaOYixcv0tvXl3WMSYzWClD09vXx'\
               b'99/4RhZCaUocx9nXmUaDOIpxIhiz8m32egLIO+emjVabrAjO'\
               b'SqtHD7CpxZgI43nQ4v+09WUmCAKSNMXh8DyfJI6m1gvAsr5d'\
               b'It3Xja7WKG3btu1ze/bsedzzPAYGBjh27Njp0dHRvwFCoNEa'\
               b'daAK1FrjI5dOWGjZzffs2SN79+59vFAoMDIyQpIkU6Ojo//x'\
               b'0Zu4snREo8tJsVhM+/v7KRQK3HHHHRSLxVXL/nrILQPwfd/2'\
               b'9/dTLBYZGBigXC7//wJQLpfT/v5+fN+nVCqhtV7/L3rLSCdJ'\
               b'zKFDh+5VSr0D9LR/qX2czoPOgUvBRigXL/quNDQ0dG7//v0j'\
               b'Sql1A6ceffTRP/F9/2MrKe3cufPeffv2/dlaF+/q6uKpp576'\
               b'uzRNl+1xlFJqYGDg4pNPPvnSWtdur3H8+PFzpVJpy0pK1Wr1'\
               b'lhbv6elZVadarZ7cvXv3jlvagA5ywFpLvV6nXl/Tpysge4/o'\
               b'6+tbVed2xBsaGmr/n8NSSZKEq1ev4nkeU1NrL6b5fJ4dO3Ys'\
               b'f/3SEpFVbglWEe/w4cMv5nK5TUv/EEWRFwRBDkBrPdTd3f3A'\
               b'Whfv7e3l6aeffp6bvJC31r6tNqMjFnrsscd2OOd+AeTXsvjO'\
               b'nTtnZ2dnBw8ePHh7cbKC/B8rHMqx5Ahr7gAAAABJRU5ErkJg'\
               b'gg=='



# -*- coding: utf-8 -*-
#
# AWL simulator - AWL optimizer - Remove NOPs
#
# Copyright 2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.awloptimizer.base import *

from awlsim.core.instructions.all_insns import * #+cimport


__all__ = [
	"AwlOptimizer_NopRemove",
]


class AwlOptimizer_NopRemove(AwlOptimizer_Base):
	"""AWL/STL optimizer: Remove NOP instructions
	"""

	NAME	= "noprem"

	def __init__(self, optimizer):
		AwlOptimizer_Base.__init__(self, optimizer)

	def run(self, insns):
		newInsns = []
		for i, insn in enumerate(insns):
			if isinstance(insn, AwlInsn_NOP):
				if insn.labelStr:
					if i < len(insns) - 1:
						# Move the label to the next insn
						nextInsn = insns[i + 1]
						if nextInsn.labelStr:
							# Rename all references to insn.labelStr
							# to nextInsn.labelStr
							pass#TODO
						else:
							# Assign the label to the next insn.
							nextInsn.labelStr = insn.labelStr
							# Remove the NOP
							continue
					else:
						# This NOP is the last insn.
						# Keep it.
						pass
				else:
					# This is just a plain NOP.
					# Remove it.
					continue
			newInsns.append(insn)
		return newInsns



# -*- coding: utf-8 -*-
#
# AWL simulator - IEC library - FC 12 "GE_DT"
#
# Copyright 2015 Christian Vitte <vitte.chris@gmail.com>
# Copyright 2015 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.library.libentry import *


class Lib__IEC__FC12_GE_DT(AwlLibFC):
	libraryName	= "IEC"
	staticIndex	= 12
	symbolName	= "GE_DT"
	description	= "Greater or equal DATE_AND_TIME"

	interfaceFields = {
		BlockInterfaceField.FTYPE_IN	: (
			BlockInterfaceField(name="DT1", dataType="DATE_AND_TIME"),
			BlockInterfaceField(name="DT2", dataType="DATE_AND_TIME"),
		),
		BlockInterfaceField.FTYPE_OUT	: (
			BlockInterfaceField(name="RET_VAL", dataType="BOOL"),
		),
		BlockInterfaceField.FTYPE_TEMP	: (
			BlockInterfaceField(name="YEAR1", dataType="WORD"),
			BlockInterfaceField(name="YEAR2", dataType="WORD"),
			BlockInterfaceField(name="AR1_SAVE", dataType="DWORD"),
		),
	}

	awlCodeCopyright = "Copyright (c) 2015 Christian Vitte <vitte.chris@gmail.com>\n"\
			   "Copyright (c) 2015 Michael Buesch <m@bues.ch>"
	awlCodeLicense = "BSD-2-clause"
	awlCodeIsStandard = True
	awlCodeVersion = "0.1"

	awlCode = """
	TAR1	#AR1_SAVE	// Save AR1 register

	// Load a pointer to #DT1 into AR1 and open the DB
	L	P##DT1
	LAR1
	L	W [AR1, P#0.0]
	T	#YEAR1 // DB number (abuse #YEAR1 temp field)
	AUF	DB [#YEAR1]
	L	D [AR1, P#2.0]
	LAR1

	// Load a pointer to #DT2 into AR2 and open the DB as DI
	L	P##DT2
	LAR2
	L	W [AR2, P#0.0]
	T	#YEAR1 // DB number (abuse #YEAR1 temp field)
	AUF	DI [#YEAR1]
	L	D [AR2, P#2.0]
	// If #DT2 points to DB (area 84) change it to DI (area 85).
	// This also works, if #DT2 points to VL (area 87).
	// Other areas are not possible.
	OD	DW#16#85000000
	LAR2

//------------------------------------------------------
	// Extract years from DT1 and DT2
	L	B [AR1, P#0.0]
	T	#YEAR1
	L	B [AR2, P#0.0]
	T	#YEAR2

	// Check whether the year values from DT1 and DT2
	// are valid BCD numbers.
	L	#YEAR1
	UW	W#16#000F
	L	W#16#0009
	>I
	SPB	FAIL
	L	#YEAR1
	UW	W#16#00F0
	L	W#16#0090
	>I
	SPB	FAIL
	L	#YEAR2
	UW	W#16#000F
	L	W#16#0009
	>I
	SPB	FAIL
	L	#YEAR2
	UW	W#16#00F0
	L	W#16#0090
	>I
	SPB	FAIL

//------------------------------------------------------
	// Check whether specified year is 1990-1999 or 2000-2089
	L	B#16#89
	L	#YEAR1
	>=I
	SPB	_2K1

	// 1900 years correction (applicable for year 1990-1999)
	OW	W#16#1900
	T	#YEAR1
	SPA	Y2

	// 2000 years correction (applicable for year 2000-2089)
_2K1:	OW	W#16#2000
	T	#YEAR1

	// Check whether specified year is 1990-1999 or 2000-2089
Y2:	L	B#16#89
	L	#YEAR2
	>=I
	SPB	_2K2

	// 1900 years correction (applicable for year 1990-1999)
	OW	W#16#1900
	T	#YEAR2
	SPA	YRCK

	// 2000 years correction (applicable for year 2000-2089)
_2K2:	OW	W#16#2000
	T	#YEAR2

//------------------------------------------------------
	// Check if YEAR1 >= YEAR2.
	// This check also works without BCD->INT conversion.
YRCK:	L	#YEAR1
	L	#YEAR2
	<I
	// year1 < year2 -> NOK
	SPB	NOK
	// year1 > year2 -> OK
	>I
	SPB	OK

//------------------------------------------------------
	// Check if M:D:H DT1 >= M:D:H DT2 - Bytes 1 to 3
	// Extract M:D:H from DT1 and DT2 (without year)
	L	D [AR1, P#0.0]
	UD	DW#16#00FFFFFF	// Remove sign (and year)
	L	D [AR2, P#0.0]
	UD	DW#16#00FFFFFF	// Remove sign (and year)
	<D
	// BCD-from-DT1 < BCD-from-DT2 -> NOK
	// This check also works without BCD->INT conversion.
	SPB	NOK
	>D
	// BCD-from-DT1 > BCD-from-DT2 -> OK
	SPB	OK

//------------------------------------------------------
	// Check if M:S:MS1 DT1 >= M:S:MS1 DT2 - Bytes 4 to 6
	// Extract M:S:MS1 from DT1 and DT2
	L	D [AR1, P#3.0]
	UD	DW#16#00FFFFFF	// Remove sign (and hour)
	L	D [AR2, P#3.0]
	UD	DW#16#00FFFFFF	// Remove sign (and hour)
	<D
	// BCD-from-DT1 < BCD-from-DT2 -> NOK
	// This check also works without BCD->INT conversion.
	SPB	NOK
	>D
	// BCD-from-DT1 > BCD-from-DT2 -> OK
	SPB	OK

//------------------------------------------------------
	// Check if MS2:WD DT1 >= MS2:WD DT2 - Byte 7
	// Extract MS2:WD from DT1 and DT2
	L	B [AR1, P#7.0]
	L	B [AR2, P#7.0]
	<I
	// BCD-from-DT1 < BCD-from-DT2 -> NOK
	// This check also works without BCD->INT conversion.
	SPB	NOK

//------------------------------------------------------
	// Everything is Ok.
	// VKE is 1.
OK:	=	#RET_VAL	// RET_VAL := 1
	SAVE			// BIE := 1
	SPA	END

	// BCD failure. Set BIE := 0.
	// Do not touch RET_VAL.
FAIL:	CLR
	SAVE			// BIE := 0
	SPA	END

	// BCD is Ok, but result is negative.
	// VKE is 1.
NOK:	SAVE			// BIE := 1
	CLR
	=	#RET_VAL	// RET_VAL := 0

END:	LAR1	#AR1_SAVE	// Restore AR1 register
	BE
"""

AwlLib.registerEntry(Lib__IEC__FC12_GE_DT)



# -*- coding: utf-8 -*-
#
# AWL simulator - Generic source code edit widget
#
# Copyright 2014 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.gui.util import *


class SourceCodeEdit(QPlainTextEdit):
	# Signal: Change the font size.
	#         If the parameter is True, increase font size.
	resizeFont = Signal(bool)

	# Signal: Validation request.
	#	  A code validation should take place.
	#	  In case of validation failure, call setErraticLine().
	#	  The parameter is self.
	validateDocument = Signal(QObject)

	def __init__(self, parent=None):
		QPlainTextEdit.__init__(self, parent)

		self.__wheelSteps = 0.0

		self.__undoIsAvailable = False
		self.__redoIsAvailable = False
		self.__copyIsAvailable = False

		self.__prevLineNr = self.textCursor().blockNumber()
		self.__prevColumnNr = self.textCursor().columnNumber()
		self.__columnChange = False

		self.enableAutoIndent()
		self.enablePasteIndent()
		self.__validateEn = True
		self.__errLineNr = None
		self.__errLineMsg = None

		self.cursorPositionChanged.connect(self.__handleCursorChange)
		self.textChanged.connect(self.__handleTextChange)
		self.undoAvailable.connect(self.__handleUndoAvailableChange)
		self.redoAvailable.connect(self.__handleRedoAvailableChange)
		self.copyAvailable.connect(self.__handleCopyAvailableChange)

	def __handleUndoAvailableChange(self, available):
		self.__undoIsAvailable = available

	def undoIsAvailable(self):
		return self.__undoIsAvailable

	def __handleRedoAvailableChange(self, available):
		self.__redoIsAvailable = available

	def redoIsAvailable(self):
		return self.__redoIsAvailable

	def __handleCopyAvailableChange(self, available):
		self.__copyIsAvailable = available

	def copyIsAvailable(self):
		return self.__copyIsAvailable

	def enableAutoIndent(self, enable=True):
		self.__autoIndentEn = enable

	def enablePasteIndent(self, enable=True):
		self.__pasteIndentEn = enable

	def enableValidation(self, enable=True):
		self.__validateEn = enable
		if not enable:
			QToolTip.hideText()
		self.setErraticLine(None)
		self.__validate()

	def __getLineIndent(self, cursor):
		cursor.select(QTextCursor.LineUnderCursor)
		line = cursor.selectedText()
		if not line:
			return ""
		# Get the indent-string (that is the whitespace line prefix)
		for i, c in enumerate(line):
			if not c.isspace():
				break
		else:
			i += 1
		return line[:i]

	def __autoIndentHandleNewline(self):
		if not self.__autoIndentEn:
			return
		# Move cursor to previous line and get its indent string.
		cursor = self.textCursor()
		if not cursor.movePosition(QTextCursor.Up,
					   QTextCursor.MoveAnchor, 1):
			return
		indentStr = self.__getLineIndent(cursor)
		# Insert the indent string into the current line
		cursor = self.textCursor()
		cursor.insertText(indentStr)
		self.setTextCursor(cursor)
		# Remove any old indent (righthand of the cursor)
		cursor = self.textCursor()
		if not cursor.movePosition(QTextCursor.EndOfLine,
					   QTextCursor.KeepAnchor, 1):
			return
		selText = cursor.selectedText()
		stripCount = len(selText) - len(selText.lstrip())
		if stripCount > 0:
			cursor = self.textCursor()
			if not cursor.movePosition(QTextCursor.Right,
						   QTextCursor.KeepAnchor,
						   stripCount):
				return
			cursor.deleteChar()

	def keyPressEvent(self, ev):
		QPlainTextEdit.keyPressEvent(self, ev)

		if ev.key() in (Qt.Key_Return, Qt.Key_Enter):
			self.__autoIndentHandleNewline()
		elif ev.key() == Qt.Key_Delete:
			self.__validate()

	def wheelEvent(self, ev):
		if ev.modifiers() & Qt.ControlModifier:
			# Ctrl + Scroll-wheel: Font resizing
			if isQt4:
				numDegrees = ev.delta() / 8
			else:
				numDegrees = ev.angleDelta().y() / 8
			numSteps = numDegrees / 15
			self.__wheelSteps += numSteps
			if self.__wheelSteps >= 1.0:
				while self.__wheelSteps >= 1.0:
					self.resizeFont.emit(False)
					self.__wheelSteps -= 1.0
			elif self.__wheelSteps <= -1.0:
				while self.__wheelSteps <= -1.0:
					self.resizeFont.emit(True)
					self.__wheelSteps += 1.0
			return
		else:
			self.__wheelSteps = 0.0

		QPlainTextEdit.wheelEvent(self, ev)

	# Add the current indent to all lines (except the first) in 'text'.
	def __makeSeamlessIndent(self, text):
		indentStr = self.__getLineIndent(self.textCursor())
		lines = []
		for i, line in enumerate(text.splitlines()):
			if i == 0:
				lines.append(line)
			else:
				lines.append(indentStr + line)
		return "\n".join(lines)

	def pasteText(self, text):
		if self.__autoIndentEn:
			text = self.__makeSeamlessIndent(text)
		self.insertPlainText(text)

	def insertFromMimeData(self, mimeData):
		if mimeData.hasText() and self.__pasteIndentEn:
			# Replace the mimeData by a re-indented mimeData.
			newText = self.__makeSeamlessIndent(mimeData.text())
			mimeData = QMimeData()
			mimeData.setText(newText)
		QPlainTextEdit.insertFromMimeData(self, mimeData)

	def __handleCursorChange(self):
		cursor = self.textCursor()
		columnNr = cursor.columnNumber()
		lineNr = cursor.blockNumber()

		if columnNr != self.__prevColumnNr:
			self.__columnChange = True

		if lineNr == self.__errLineNr or\
		   (lineNr != self.__prevLineNr and\
		    self.__columnChange):
			self.__validate()
			self.__columnChange = False

		self.__prevColumnNr = columnNr
		self.__prevLineNr = lineNr

	def __handleTextChange(self):
		QToolTip.hideText()
		if self.__errLineNr is not None:
			self.__validate()

	def setPlainText(self, text):
		self.__errLineNr = None
		self.__errLineMsg = None
		QPlainTextEdit.setPlainText(self, text)
		self.__validate()

	@staticmethod
	def __makeExtraSel(cursor, format):
		sel = QTextEdit.ExtraSelection()
		sel.cursor = cursor
		sel.format = format
		return sel

	@staticmethod
	def __makeTextCursorLineSel(cursor, lineNr):
		cursor.movePosition(QTextCursor.Start,
				    QTextCursor.MoveAnchor, 1)
		cursor.movePosition(QTextCursor.Down,
				    QTextCursor.MoveAnchor, lineNr)
		cursor.select(QTextCursor.LineUnderCursor)
		return cursor

	__errLineBrush = QBrush(getErrorColor())

	def __validate(self):
		if not self.__validateEn:
			return

		# Validate the current document
		text = self.toPlainText()
		if text.strip():
			# Run the validator.
			self.validateDocument.emit(self)
		else:
			# No text. Mark everything as ok.
			self.setErraticLine(None)

	def __tryShowValidateErrToolTip(self, lineNr, globalPos):
		errLineNr = self.__errLineNr
		errLineMsg = self.__errLineMsg
		if self.__validateEn and\
		   not self.textCursor().hasSelection() and\
		   errLineNr is not None and\
		   errLineMsg:
			errLineNr += 1
			if lineNr in {errLineNr - 1, errLineNr, errLineNr + 1}:
				lines = []
				for line in errLineMsg.splitlines():
					# Limit line length.
					if len(line) > 237:
						line = line[:237] + "..."
					# Enforce line breaks.
					while len(line) > 80:
						lines.append(line[:80] + " \\")
						line = line[80:]
					lines.append(line)
				errLineMsg = "\n".join(lines)
				# Show the validation result.
				QToolTip.showText(globalPos, errLineMsg,
						  self, QRect())
				return True
		return False

	# Mark an erratic line
	def setErraticLine(self, lineNr, errMsg=None):
		if lineNr is None:
			self.setExtraSelections(())
		else:
			cursor = self.textCursor()
			fmt = self.currentCharFormat()
			fmt.setBackground(self.__errLineBrush)
			extraSel = self.__makeExtraSel(
				self.__makeTextCursorLineSel(cursor, lineNr),
				fmt)
			self.setExtraSelections((extraSel,))

		self.__errLineNr = lineNr
		self.__errLineMsg = errMsg

		if errMsg:
			mouseCursorPos = QCursor.pos()
			cursor = self.cursorForPosition(self.mapFromGlobal(mouseCursorPos))
			self.__tryShowValidateErrToolTip(cursor.blockNumber(),
							 mouseCursorPos)

	def __toolTipEvent(self, ev):
		cursor = self.cursorForPosition(ev.pos())
		if not self.__tryShowValidateErrToolTip(cursor.blockNumber(),
							ev.globalPos()):
			QToolTip.hideText()

	def event(self, ev):
		if ev.type() == QEvent.ToolTip:
			self.__toolTipEvent(ev)
		return super(SourceCodeEdit, self).event(ev)



# -*- coding: utf-8 -*-
#
# AWL simulator - Hardware module parameters
#
# Copyright 2013-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.exceptions import *


__all__ = [ "HwParamDesc",
	    "HwParamDesc_pyobject",
	    "HwParamDesc_str",
	    "HwParamDesc_int",
	    "HwParamDesc_bool",
]


class HwParamDesc(object):
	"""Abstract hardware parameter descriptor."""

	class ParseError(Exception):
		pass

	typeStr = "<NoType>"
	userEditable = True
	defaultValue = None

	def __init__(self, name, description="", mandatory=False):
		self.name = name
		self.description = description
		self.mandatory = mandatory

	def parse(self, value):
		"""Parse a value string.
		This must be overridden.
		"""
		raise NotImplementedError

	def match(self, matchName):
		"""Match a name string.
		The default implementation just compares the string to self.name.
		The parameter must not be 'mandatory', if this method is overridden.
		"""
		return self.name == matchName

class HwParamDesc_pyobject(HwParamDesc):
	"""Generic object parameter descriptor."""

	typeStr = "PyObject"
	userEditable = False

	def __init__(self, name, description="", mandatory=False):
		HwParamDesc.__init__(self, name, description, mandatory)

	def parse(self, value):
		return value

class HwParamDesc_str(HwParamDesc):
	"""String hardware parameter descriptor."""

	typeStr = "string"

	def __init__(self, name, defaultValue="", description="", mandatory=False):
		HwParamDesc.__init__(self, name, description, mandatory)
		self.defaultValue = defaultValue

	def parse(self, value):
		return value

class HwParamDesc_int(HwParamDesc):
	"""Integer hardware parameter descriptor."""

	typeStr = "integer"

	def __init__(self, name,
		     defaultValue=0, minValue=None, maxValue=None,
		     description="", mandatory=False):
		HwParamDesc.__init__(self, name, description, mandatory)
		self.defaultValue = defaultValue
		self.minValue = minValue
		self.maxValue = maxValue

	def parse(self, value):
		try:
			value = int(value)
		except ValueError:
			raise self.ParseError("Value '%s' is not a valid integer." %\
					      str(value))
		if self.minValue is not None:
			if value < self.minValue:
				raise self.ParseError("Value '%d' is too small." % value)
		if self.maxValue is not None:
			if value > self.maxValue:
				raise self.ParseError("Value '%d' is too big." % value)
		return value

class HwParamDesc_bool(HwParamDesc):
	"""Boolean hardware parameter descriptor."""

	typeStr = "boolean"

	def __init__(self, name, defaultValue=False,
		     description="", mandatory=False):
		HwParamDesc.__init__(self, name, description, mandatory)
		self.defaultValue = defaultValue

	def parse(self, value):
		value = value.strip()
		if value.lower() in ("true", "yes", "on"):
			return True
		if value.lower() in ("false", "no", "off"):
			return False
		try:
			value = int(value, 10)
		except ValueError:
			raise self.ParseError("Value '%s' is not a valid boolean." %\
				str(value))
		return bool(value)



# -*- coding: utf-8 -*-
#
# PiLC Raspberry Pi HAT configuration library
#
# Copyright 2016 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals

import math
import smbus


class PilcConf(object):
	"""Hardware configuration of the PiLC hat.
	"""

	DEFAULT_BUS	= 0x00
	DEFAULT_DEV	= 0x2F

	CONF_NONE	= 0
	CONF_XTALCAL	= 1
	CONF_EEMUWE	= 2
	CONF_PBTXENDBG	= 3
	CONF_PBTXENTO	= 4

	sizes = {
		CONF_NONE	: 0,
		CONF_XTALCAL	: 8,
		CONF_EEMUWE	: 1,
		CONF_PBTXENDBG	: 8,
		CONF_PBTXENTO	: 16,
	}

	PBTXEN_DBG_OFF		= 0
	PBTXEN_DBG_RETRIG	= 1
	PBTXEN_DBG_NOTRIG	= 2

	TRIES = 5

	class Error(Exception):
		pass

	@classmethod
	def havePilcHat(cls):
		"""Returns True, if a PiLC HAT is attached to the device.
		"""
		try:
			with open("/proc/device-tree/hat/product", "rb") as fd:
				prod = fd.read().decode("UTF-8").rstrip('\0').strip()
				if prod != "PiLC":
					return False
		except (IOError, UnicodeError) as e:
			return False
		return True

	def __init__(self, bus = DEFAULT_BUS, dev = DEFAULT_DEV):
		try:
			self.__i2c = smbus.SMBus()
			self.__i2c.open(bus)
			self.__dev = dev
		except (OSError, IOError) as e:
			raise self.Error("Failed to init I2C communication:\n" + str(e))

	def close(self):
		try:
			return self.__i2c.close()
		except (OSError, IOError) as e:
			raise self.Error("Failed to close I2C communication:\n" + str(e))

	def __baudToFrameUs(self, kBaud):
		"""Get the frame length in microseconds.
		"""
		symsPerFrame = 1 + 8 + 1 + 1
		symUs = (1.0 / (kBaud * 1000)) * 1000000
		frameUs = symUs * symsPerFrame
		return int(math.ceil(frameUs))

	def setBaudrate(self, kBaud):
		self.set(self.CONF_PBTXENTO, self.__baudToFrameUs(kBaud))

	def get(self, confItem):
		itemSize = self.sizes[confItem]
		handlers = {
			0	: lambda confItem: None,
			1	: self.__readBool,
			8	: self.__readU8,
			16	: self.__readU16,
		}
		for i in range(self.TRIES):
			try:
				return handlers[itemSize](confItem)
			except (OSError, IOError) as e:
				continue
		raise self.Error("Failed to read config item %d" % confItem)

	def set(self, confItem, value):
		itemSize = self.sizes[confItem]
		handlers = {
			0	: lambda confItem, data: None,
			1	: self.__writeBool,
			8	: self.__writeU8,
			16	: self.__writeU16,
		}
		for i in range(self.TRIES):
			try:
				handlers[itemSize](confItem, value)
				checkValue = self.get(confItem)
				if value == checkValue:
					return
			except (OSError, IOError) as e:
				continue
		raise self.Error("Failed to write config item %d" % confItem)

	def __writeBool(self, confItem, data):
		data = 1 if data else 0
		self.__writeU8(confItem, data)

	def __writeU8(self, confItem, data):
		payload = [ (data & 0xFF), (~data & 0xFF), ]
		self.__i2c.write_i2c_block_data(self.__dev, confItem, payload)

	def __writeU16(self, confItem, data):
		payload = [ (data & 0xFF), ((data >> 8) & 0xFF),
			    (~data & 0xFF), ((~data >> 8) & 0xFF), ]
		self.__i2c.write_i2c_block_data(self.__dev, confItem, payload)

	def __readBool(self, confItem):
		return True if self.__readU8(confItem) else False

	def __readU8(self, confItem):
		payload = self.__i2c.read_i2c_block_data(self.__dev, confItem, 1)
		return payload[0]

	def __readU16(self, confItem):
		payload = self.__i2c.read_i2c_block_data(self.__dev, confItem, 2)
		return payload[0] | (payload[1] << 8)
